#include <aftermath/trace/trace.h>
#include <stdlib.h>

/* This very simple example creates a trave with a certain number of workers,
 * each traversing different run-time states. The example defines three states:
 * init (initialization of the worker), busy (the worker performs some useful
 * work) and term (worker is terminating).
 *
 * In contrast to states.c, the numerical IDs of states are not consecutive.
 */

enum {
	STATE_INIT = 0,
	STATE_BUSY = 276876922,
	STATE_TERM = 17899992
};

/* Functions emulating a worker. In a real example, these would be worker
 * threads. */
void worker_fun1(struct am_event_set* es);
void worker_fun2(struct am_event_set* es);

int main(int argc, char** argv)
{
	/* Object representing an entire trace */
	struct am_trace trace;

	/* Each worker has its own, private event set that it traces events
	 * into. We'll create the two event sets implicitly by registering CPUs
	 * and then retrieve a pointer to a CPU's event set for each worker. */
	struct am_event_set* worker_1;
	struct am_event_set* worker_2;

	/* Size of the global buffer storing trace global events and some
	 * metadata. This buffer can be fairly small, since most of the time
	 * there are only very few global events. */
	size_t global_buffer_size = 64*1024;

	/* The size of the buffer for a worker. These are the buffers receiving
	 * most of the data, so they should be relatively big, depending on the
	 * workload of course. Libaftermath-trace allocates these buffers
	 * through calls of libc's malloc() function. On modern systems, this
	 * allocates the buffers only logically and pages are allocated
	 * physically as the buffer is filled with data. Hence, in most cases,
	 * overprovisioning does not actually increase the memory footprint of
	 * the instrumented run-time. */
	size_t per_worker_buffer_size = 1024*1024;

	/* Output file */
	const char* trace_filename;

	if(argc != 2) {
		fprintf(stderr, "Usage: %s output_file\n", argv[0]);
		exit(1);
	}

	trace_filename = argv[1];

	/* Initialize the trace object */
	if(am_trace_init(&trace, global_buffer_size)) {
		fprintf(stderr, "Could not initialize trace\n");
		exit(1);
	}

	/* Register states. For each registered state, a state description
	 * associating the numerical ID of the state (e.g., STATE_INIT) with the
	 * human-readable name (e.g., "Initialization") is generated and
	 * appended to the trace. */
	if(am_trace_register_state(&trace, STATE_INIT, "Initialization") ||
	   am_trace_register_state(&trace, STATE_BUSY, "Busy") ||
	   am_trace_register_state(&trace, STATE_TERM, "Termination"))
	{
		fprintf(stderr, "Could not register states\n");
		exit(1);
	}

	/* Indicate that we will trace events for two CPUs: CPU 6 and CPU 9
	 * (chosen totally abitrarily). */
	if(am_trace_register_cpu(&trace, 6, per_worker_buffer_size) ||
	   am_trace_register_cpu(&trace, 9, per_worker_buffer_size)) {
		fprintf(stderr, "Could not initialize event set\n");
		exit(1);
	}

	/* Retrieve pointer to event set for CPU 6 and CPU 9. We'll pretend that
	 * the first worker executes on CPU 6 and the second worker on CPU 9. */
	worker_1 = am_trace_get_event_set(&trace, 6);
	worker_2 = am_trace_get_event_set(&trace, 9);

	/* End of the initialization. In a real run-time execution would be
	 * concurrent after this point and the workers would trace events
	 * independently. In this example, we'll just execute two sequential
	 * functions emulating the workers' behavior. */
	worker_fun1(worker_1);
	worker_fun2(worker_2);

	/* End of execution. A real run-time would synchronize would wait for
	 * the workers to terminate to ensure that no events are traced while
	 * the trace is written to disk. */

	/* Write the contents of the trace to disk. */
	if(am_trace_dump(&trace, trace_filename)) {
		fprintf(stderr, "Could not write trace to file \"%s\"\n",
			trace_filename);
		exit(1);
	}

	/* De-allocate trace, destroy all event sets, buffers, etc. */
	am_trace_destroy(&trace);

	return 0;
}

void worker_fun1(struct am_event_set* es)
{
	/* Trace that this worker is in the initialization state from 0 to 1000
	 * cycles */
	if(am_event_set_trace_state(es, STATE_INIT, 0, 1000)) {
		fprintf(stderr, "Could not write trace state.\n");
		exit(1);
	}

	/* Indicate busy period from 1215 to 9103 cycles */
	if(am_event_set_trace_state(es, STATE_BUSY, 1215, 9103)) {
		fprintf(stderr, "Could not write trace state.\n");
		exit(1);
	}

	/* Indicate another busy period from 18902 to 22005 cycles */
	if(am_event_set_trace_state(es, STATE_BUSY, 18902, 22005)) {
		fprintf(stderr, "Could not write trace state.\n");
		exit(1);
	}

	/* Termination from 22010 to 23000 cycles */
	if(am_event_set_trace_state(es, STATE_TERM, 22010, 23000)) {
		fprintf(stderr, "Could not write trace state.\n");
		exit(1);
	}
}

void worker_fun2(struct am_event_set* es)
{
	/* Trace that this worker is in the initialization state from 105 to 995
	 * cycles */
	if(am_event_set_trace_state(es, STATE_INIT, 105, 995)) {
		fprintf(stderr, "Could not write trace state.\n");
		exit(1);
	}

	/* A single busy period from 2000 to 8000 cycles */
	if(am_event_set_trace_state(es, STATE_BUSY, 2000, 8000)) {
		fprintf(stderr, "Could not write trace state.\n");
		exit(1);
	}

	/* Termination from 8222 to 9001 cycles */
	if(am_event_set_trace_state(es, STATE_TERM, 8222, 9001)) {
		fprintf(stderr, "Could not write trace state.\n");
		exit(1);
	}
}
